@@ -56,15 +56,21 @@ An [aryref] of other aliases for the same command.  optional.
 
 The type of this setting:
 
-           - $CONF_TYPE_STRING: string
-           - $CONF_TYPE_NUMERIC: numeric value (float or int)
-           - $CONF_TYPE_BOOL: boolean (0 or 1)
-           - $CONF_TYPE_TEMPLATE: template, like "report"
-           - $CONF_TYPE_ADDRLIST: address list, like "whitelist_from"
-           - $CONF_TYPE_HASH_KEY_VALUE: hash key/value pair,
-             like "describe" or tflags
+ - $CONF_TYPE_NOARGS: must not have any argument, like "clear_headers"
+ - $CONF_TYPE_STRING: string
+ - $CONF_TYPE_NUMERIC: numeric value (float or int)
+ - $CONF_TYPE_BOOL: boolean (0/no or 1/yes)
+ - $CONF_TYPE_TEMPLATE: template, like "report"
+ - $CONF_TYPE_ADDRLIST: list of mail addresses, like "whitelist_from"
+ - $CONF_TYPE_HASH_KEY_VALUE: hash key/value pair, like "describe" or tflags
+ - $CONF_TYPE_STRINGLIST list of strings, stored as an array
+ - $CONF_TYPE_IPADDRLIST list of IP addresses, stored as an array of SA::NetSet
 
-If this is set, a 'code' block is assigned based on the type.
+If this is set, and a 'code' block does not already exist, a 'code' block is
+assigned based on the type.
+
+In addition, the SpamAssassin test suite will validate that the settings
+do not 'leak' between users.
 
 Note that C<$CONF_TYPE_HASH_KEY_VALUE>-type settings require that the
 value be non-empty, otherwise they'll produce a warning message.
@@ -88,6 +94,10 @@ from a set of 'valid' values, but the user provided an invalid one.
 Any other values -- including C<undef> -- returned from the subroutine are
 considered to mean 'success'.
 
+It is good practice to set a 'type', if possible, describing how your settings
+are stored on the Conf object; this allows the SpamAssassin test suite to
+validate that the settings do not 'leak' between users.
+
 =item default
 
 The default value for the setting.  may be omitted if the default value is a
@@ -124,10 +134,13 @@ package Mail::SpamAssassin::Conf::Parser;
 use Mail::SpamAssassin::Conf;
 use Mail::SpamAssassin::Constants qw(:sa);
 use Mail::SpamAssassin::Logger;
+use Mail::SpamAssassin::Util qw(untaint_var);
+use Mail::SpamAssassin::NetSet;
 
 use strict;
 use warnings;
 use bytes;
+use re 'taint';
 
 use vars qw{
   @ISA
@@ -191,8 +204,12 @@ sub build_command_luts {
     # next, its priority (used to ensure frequently-used params
     # are parsed first)
     my $cmdname = $cmd->{command} || $cmd->{setting};
-    foreach my $name ($cmdname, @{$cmd->{aliases}}) {
-      $self->{command_luts}->{$set}->{$name} = $cmd;
+    $self->{command_luts}->{$set}->{$cmdname} = $cmd;
+
+    if ($cmd->{aliases} && scalar @{$cmd->{aliases}} > 0) {
+      foreach my $name (@{$cmd->{aliases}}) {
+        $self->{command_luts}->{$set}->{$name} = $cmd;
+      }
     }
   }
 }
@@ -231,14 +248,24 @@ sub parse {
 
   $self->{currentfile} = '(no file)';
   my $skip_parsing = 0;
-  my @curfile_stack = ();
-  my @if_stack = ();
+  my @curfile_stack;
+  my @if_stack;
   my @conf_lines = split (/\n/, $_[1]);
   my $line;
+  $self->{if_stack} = \@if_stack;
+  $self->{file_scoped_attrs} = { };
+
+  my $keepmetadata = $conf->{main}->{keep_config_parsing_metadata};
 
   while (defined ($line = shift @conf_lines)) {
     local ($1);         # bug 3838: prevent random taint flagging of $1
 
+    # bug 5545: used to support testing rules in the ruleqa system
+    if ($keepmetadata && $line =~ /^\#testrules/) {
+      $self->{file_scoped_attrs}->{testrules}++;
+      next;
+    }
+
     $line =~ s/(?<!\\)#.*$//; # remove comments
     $line =~ s/\\#/#/g; # hash chars are escaped, so unescape them
     $line =~ s/^\s+//;  # remove leading whitespace
@@ -252,11 +279,11 @@ sub parse {
     $key = lc $key;
     # convert all dashes in setting name to underscores.
     $key =~ s/-/_/g;
-
-    # Do a better job untainting this info ...
     $value = '' unless defined($value);
-    $value =~ /^(.*)$/;
-    $value = $1;
+
+#   # Do a better job untainting this info ...
+#   # $value = untaint_var($value);
+#   Do NOT blindly untaint now, do it carefully later when semantics is known!
 
     my $parse_error;       # undef by default, may be overridden
 
@@ -269,12 +296,16 @@ sub parse {
       }
 
       if ($value =~ /^end\s/) {
+        $self->{file_scoped_attrs} = { };
+
         if (scalar @if_stack > 0) {
           my $cond = pop @if_stack;
 
           if ($cond->{type} eq 'if') {
-            warn "config: unclosed 'if' in ".
+            my $msg = "config: unclosed 'if' in ".
                   $self->{currentfile}.": if ".$cond->{conditional}."\n";
+            warn $msg;
+            $self->lint_warn($msg, undef);
           }
           else {
             # die seems a bit excessive here, but this shouldn't be possible
@@ -282,7 +313,6 @@ sub parse {
             die "config: unknown 'if' type: ".$cond->{type}."\n";
           }
 
-          $conf->{errors}++;
           @if_stack = ();
         }
         $skip_parsing = 0;
@@ -317,9 +347,25 @@ sub parse {
       next;
     }
 
+    if ($key eq 'else') {
+      # TODO: if/else/else won't get flagged here :(
+      if (!@if_stack) {
+        $parse_error = "config: found else without matching conditional";
+        goto failed_line;
+      }
+
+      $skip_parsing = !$skip_parsing;
+      next;
+    }
+
     # and the endif statement:
     if ($key eq 'endif') {
       my $lastcond = pop @if_stack;
+      if (!defined $lastcond) {
+        $parse_error = "config: found endif without matching conditional";
+        goto failed_line;
+      }
+
       $skip_parsing = $lastcond->{skip_parsing};
       next;
     }
@@ -342,13 +388,14 @@ sub parse {
       #$value =~ s/^(\d+)\.(\d{1,3}).*$/sprintf "%d.%d", $1, $2/e;
 
       if ($ver ne $value) {
-        warn "config: configuration file \"$self->{currentfile}\" requires version ".
-                "$value of SpamAssassin, but this is code version ".
+        my $msg = "config: configuration file \"$self->{currentfile}\" requires ".
+                "version $value of SpamAssassin, but this is code version ".
                 "$ver. Maybe you need to use ".
                 "the -C switch, or remove the old config files? ".
                 "Skipping this file";
+        warn $msg;
+        $self->lint_warn($msg, undef);
         $skip_parsing = 1;
-        $conf->{errors}++;
       }
       next;
     }
@@ -420,24 +467,18 @@ failed_line:
       if ($migrated_keys{$key}) {
         # this key was moved into a plugin; non-fatal for lint
         $is_error = 0;
-        $msg = "config: failed to parse, now a plugin, skipping: $line";
+        $msg = "config: failed to parse, now a plugin, skipping, in \"$self->{currentfile}\": $line";
       } else {
         # a real syntax error; this is fatal for --lint
-        $msg = "config: failed to parse line, skipping: $line";
+        $msg = "config: failed to parse line, skipping, in \"$self->{currentfile}\": $line";
       }
     }
 
-    if ($conf->{lint_rules}) {
-      warn $msg."\n";
-    } else {
-      info($msg);
-    }
-
-    if ($is_error) {
-      $conf->{errors}++;
-    }
+    $self->lint_warn($msg, undef, $is_error);
   }
 
+  delete $self->{if_stack};
+
   $self->lint_check();
   $self->set_default_scores();
 
@@ -454,18 +495,22 @@ sub handle_conditional {
   my $eval = '';
   my $bad = 0;
   foreach my $token (@tokens) {
-    if ($token =~ /^(\W+|[+-]?\d+(?:\.\d+)?)$/) {
-      $eval .= $1." ";          # note: untaints!
+    if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/) {
+      $eval .= untaint_var($token) . " ";
     }
     elsif ($token eq 'plugin') {
       # replace with method call
       $eval .= "\$self->cond_clause_plugin_loaded";
     }
+    elsif ($token eq 'can') {
+      # replace with method call
+      $eval .= "\$self->cond_clause_can";
+    }
     elsif ($token eq 'version') {
       $eval .= $Mail::SpamAssassin::VERSION." ";
     }
-    elsif ($token =~ /^(\w[\w\:]+)$/) { # class name
-      $eval .= "\"$1\" ";       # note: untaints!
+    elsif ($token =~ /^\w[\w\:]+$/) { # class name
+      $eval .= '"' . untaint_var($token) . '" ';
     }
     else {
       $bad++;
@@ -474,7 +519,7 @@ sub handle_conditional {
   }
 
   if ($bad) {
-    $conf->{errors}++;
+    $self->lint_warn("bad 'if' line, in \"$self->{currentfile}\"", undef);
     return -1;
   }
 
@@ -488,6 +533,7 @@ sub handle_conditional {
     # leave $skip_parsing as-is; we may not be parsing anyway in this block.
     # in other words, support nested 'if's and 'require_version's
   } else {
+    warn "config: error in $key - $eval: $@" if $@ ne '';
     $$skip_parsing_ref = 1;
   }
 }
@@ -497,6 +543,17 @@ sub cond_clause_plugin_loaded {
   return $_[0]->{conf}->{plugins_loaded}->{$_[1]};
 }
 
+sub cond_clause_can {
+  my ($self, $method) = @_;
+
+  if ($method =~ /^(.*)::([^:]+)$/) {
+    return UNIVERSAL::can($1, $2);
+  } else {
+    $self->lint_warn("bad 'if' line, cannot find '::' in can($method), ".
+                "in \"$self->{currentfile}\"", undef);
+  }
+}
+
 # Let's do some linting here ...
 # This is called from _parse(), BTW, so we can check for $conf->{tests}
 # easily before finish_parsing() is called and deletes it.
@@ -504,22 +561,19 @@ sub cond_clause_plugin_loaded {
 sub lint_check {
   my ($self) = @_;
   my $conf = $self->{conf};
-  my ($k, $v);
 
-  if ($conf->{lint_rules})
-  {
+  if ($conf->{lint_rules}) {
     # Check for description and score issues in lint fashion
-    while ( ($k,$v) = each %{$conf->{descriptions}} ) {
+    while ( my $k = each %{$conf->{descriptions}} ) {
       if (!exists $conf->{tests}->{$k}) {
-        warn "config: warning: description exists for non-existent rule $k\n";
-        $conf->{errors}++;
+        $self->lint_warn("config: warning: description exists for non-existent rule $k\n", $k);
       }
     }
 
     while ( my($sk) = each %{$conf->{scores}} ) {
       if (!exists $conf->{tests}->{$sk}) {
-        warn "config: warning: score set for non-existent rule $sk\n";
-        $conf->{errors}++;
+        # bug 5514: not a lint warning any more
+        dbg("config: warning: score set for non-existent rule $sk");
       }
     }
   }
@@ -533,14 +587,13 @@ sub lint_check {
 sub set_default_scores {
   my ($self) = @_;
   my $conf = $self->{conf};
-  my ($k, $v);
 
-  while ( ($k,$v) = each %{$conf->{tests}} ) {
+  while ( my $k = each %{$conf->{tests}} ) {
     if ( ! exists $conf->{scores}->{$k} ) {
       # T_ rules (in a testing probationary period) get low, low scores
       my $set_score = ($k =~/^T_/) ? 0.01 : 1.0;
 
-      $set_score = -$set_score if ( $conf->{tflags}->{$k} =~ /\bnice\b/ );
+      $set_score = -$set_score if ( ($conf->{tflags}->{$k}||'') =~ /\bnice\b/ );
       for my $index (0..3) {
         $conf->{scoreset}->[$index]->{$k} = $set_score;
       }
@@ -572,6 +625,15 @@ sub setup_default_code_cb {
   elsif ($type == $Mail::SpamAssassin::Conf::CONF_TYPE_TEMPLATE) {
     $cmd->{code} = \&set_template_append;
   }
+  elsif ($type == $Mail::SpamAssassin::Conf::CONF_TYPE_NOARGS) {
+    $cmd->{code} = \&set_no_value;
+  }
+  elsif ($type == $Mail::SpamAssassin::Conf::CONF_TYPE_STRINGLIST) {
+    $cmd->{code} = \&set_string_list;
+  }
+  elsif ($type == $Mail::SpamAssassin::Conf::CONF_TYPE_IPADDRLIST) {
+    $cmd->{code} = \&set_ipaddr_list;
+  }
   else {
     warn "config: unknown conf type $type!";
     return 0;
@@ -579,6 +641,14 @@ sub setup_default_code_cb {
   return 1;
 }
 
+sub set_no_value {
+  my ($conf, $key, $value, $line) = @_;
+
+  unless (!defined $value || $value eq '') {
+    return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+  }
+}
+
 sub set_numeric_value {
   my ($conf, $key, $value, $line) = @_;
 
@@ -588,8 +658,8 @@ sub set_numeric_value {
   unless ($value =~ /^-?\d+(?:\.\d+)?$/) {
     return $Mail::SpamAssassin::Conf::INVALID_VALUE;
   }
-
-  $conf->{$key} = $value+0.0;
+  # it is safe to untaint now that we now the syntax is a valid number
+  $conf->{$key} = untaint_var($value) + 0.0;
 }
 
 sub set_bool_value {
@@ -598,11 +668,20 @@ sub set_bool_value {
   unless (defined $value && $value !~ /^$/) {
     return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
   }
-  unless ($value =~ /^[01]$/) {
+
+  # bug 4462: allow yes/1 and no/0 for boolean values
+  $value = lc $value;
+  if ($value eq 'yes' || $value eq '1') {
+    $value = 1;
+  }
+  elsif ($value eq 'no' || $value eq '0') {
+    $value = 0;
+  }
+  else {
     return $Mail::SpamAssassin::Conf::INVALID_VALUE;
   }
 
-  $conf->{$key} = $value+0;
+  $conf->{$key} = $value;
 }
 
 sub set_string_value {
@@ -612,7 +691,30 @@ sub set_string_value {
     return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
   }
 
-  $conf->{$key} = $value;
+  $conf->{$key} = $value;  # keep tainted
+}
+
+sub set_string_list {
+  my ($conf, $key, $value, $line) = @_;
+
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
+
+  push(@{$conf->{$key}}, split(' ', $value));
+}
+
+sub set_ipaddr_list {
+  my ($conf, $key, $value, $line) = @_;
+
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
+
+  foreach my $net (split(' ', $value)) {
+    $conf->{$key}->add_cidr($net);
+  }
+  $conf->{$key.'_configured'} = 1;
 }
 
 sub set_hash_key_value {
@@ -623,7 +725,7 @@ sub set_hash_key_value {
     return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
   }
 
-  $conf->{$key}->{$k} = $v;
+  $conf->{$key}->{$k} = $v;  # keep tainted
 }
 
 sub set_addrlist_value {
@@ -632,7 +734,7 @@ sub set_addrlist_value {
   unless (defined $value && $value !~ /^$/) {
     return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
   }
-  $conf->{parser}->add_to_addrlist ($key, split (' ', $value));
+  $conf->{parser}->add_to_addrlist ($key, split (' ', $value));  # keep tainted
 }
 
 sub remove_addrlist_value {
@@ -647,20 +749,42 @@ sub remove_addrlist_value {
 sub set_template_append {
   my ($conf, $key, $value, $line) = @_;
   if ( $value =~ /^"(.*?)"$/ ) { $value = $1; }
-  $conf->{$key} .= $value."\n";
+  $conf->{$key} .= $value."\n";  # keep tainted
 }
 
 sub set_template_clear {
   my ($conf, $key, $value, $line) = @_;
+  unless (!defined $value || $value eq '') {
+    return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+  }
   $conf->{$key} = '';
 }
 
 ###########################################################################
 
 sub finish_parsing {
-  my ($self) = @_;
+  my ($self, $isuserconf) = @_;
   my $conf = $self->{conf};
 
+  # note: this function is called once for system-wide configuration
+  # with $isuserconf set to 0, then again for user conf with $isuserconf set to 1.
+  if (!$isuserconf) {
+    $conf->{main}->call_plugins("finish_parsing_start", { conf => $conf });
+  } else {
+    $conf->{main}->call_plugins("user_conf_parsing_start", { conf => $conf });
+  }
+
+  $self->trace_meta_dependencies();
+  $self->fix_priorities();
+
+  # don't do this if allow_user_rules is active, since it deletes entries
+  # from {tests}
+  if (!$conf->{allow_user_rules}) {
+    $self->find_dup_rules();          # must be after fix_priorities()
+  }
+
+  dbg("config: finish parsing");
+
   while (my ($name, $text) = each %{$conf->{tests}}) {
     my $type = $conf->{test_types}->{$name};
     my $priority = $conf->{priority}->{$name} || 0;
@@ -668,53 +792,39 @@ sub finish_parsing {
 
     # eval type handling
     if (($type & 1) == 1) {
-      my @args;
       if (my ($function, $args) = ($text =~ m/(.*?)\s*\((.*?)\)\s*$/)) {
-	if ($args) {
-	  # bug 4419: Parse quoted strings, unquoted alphanumerics/floats and
-	  # both unquoted IPv4 and IPv6 addresses.  s// is used so that we can
-	  # determine whether or not we successfully parsed ALL arguments.
-	  while ($args =~ s/^\s*(?:['"](.*?)['"]|([\d\.:A-Za-z]+?))\s*(?:,\s*|$)//) {
-	    if (defined $1) {
-	      push @args, $1;
-	    }
-	    else {
-	      push @args, $2;
-	    }
-	  }
-	}
-        unshift(@args, $function);
-	if ($args) {
-	  $conf->{errors}++;
-	  warn("syntax error (unparsable argument: $args) for eval function: $name: $text");
-	}
+        my ($packed, $argsref) =
+                $self->pack_eval_method($function, $args, $name, $text);
+
+        if (!$packed) {
+          # we've already warned about this
+        }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS) {
-          $conf->{body_evals}->{$priority}->{$name} = \@args;
+          $conf->{body_evals}->{$priority}->{$name} = $packed;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS) {
-          $conf->{head_evals}->{$priority}->{$name} = \@args;
+          $conf->{head_evals}->{$priority}->{$name} = $packed;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS) {
           # We don't do priorities for $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS
-          $conf->{rbl_evals}->{$name} = \@args;
+          # we also use the arrayref instead of the packed string
+          $conf->{rbl_evals}->{$name} = [ $function, @$argsref ];
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_RAWBODY_EVALS) {
-          $conf->{rawbody_evals}->{$priority}->{$name} = \@args;
+          $conf->{rawbody_evals}->{$priority}->{$name} = $packed;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_FULL_EVALS) {
-          $conf->{full_evals}->{$priority}->{$name} = \@args;
+          $conf->{full_evals}->{$priority}->{$name} = $packed;
         }
         #elsif ($type == $Mail::SpamAssassin::Conf::TYPE_URI_EVALS) {
-        #  $conf->{uri_evals}->{$priority}->{$name} = \@args;
+        #  $conf->{uri_evals}->{$priority}->{$name} = $packed;
         #}
         else {
-          $conf->{errors}++;
-	  warn("unknown type $type for $name: $text");
+          $self->lint_warn("unknown type $type for $name: $text", $name);
         }
       }
       else {
-        $conf->{errors}++;
-        warn("syntax error for eval function $name: $text");
+        $self->lint_warn("syntax error for eval function $name: $text", $name);
       }
     }
     # non-eval tests
@@ -726,15 +836,6 @@ sub finish_parsing {
         $conf->{head_tests}->{$priority}->{$name} = $text;
       }
       elsif ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS) {
-        # Meta Tests must have a priority of at least META_TEST_MIN_PRIORITY,
-        # if it's lower then reset the value
-        if ($priority < META_TEST_MIN_PRIORITY) {
-          # we need to lower the count of the old priority and raise the
-          # count of the new priority
-          $conf->{priorities}->{$priority}--;
-          $priority = META_TEST_MIN_PRIORITY;
-          $conf->{priorities}->{$priority}++;
-        }
         $conf->{meta_tests}->{$priority}->{$name} = $text;
       }
       elsif ($type == $Mail::SpamAssassin::Conf::TYPE_URI_TESTS) {
@@ -746,20 +847,192 @@ sub finish_parsing {
       elsif ($type == $Mail::SpamAssassin::Conf::TYPE_FULL_TESTS) {
         $conf->{full_tests}->{$priority}->{$name} = $text;
       }
+      elsif ($type == $Mail::SpamAssassin::Conf::TYPE_EMPTY_TESTS) {
+      }
       else {
-        $conf->{errors}++;
-        warn("unknown type $type for $name: $text");
+        $self->lint_warn("unknown type $type for $name: $text", $name);
       }
     }
   }
 
   $self->lint_trusted_networks();
 
-  # named this way just in case we ever want a "finish_parsing_start"
-  $conf->{main}->call_plugins("finish_parsing_end", { conf => $conf });
+  if (!$isuserconf) {
+    $conf->{main}->call_plugins("finish_parsing_end", { conf => $conf });
+  } else {
+    $conf->{main}->call_plugins("user_conf_parsing_end", { conf => $conf });
+  }
 
-  delete $conf->{tests};                # free it up
-  delete $conf->{priority};             # free it up
+  $conf->found_any_rules();     # before we might delete {tests}
+
+  if (!$conf->{allow_user_rules}) {
+    # free up stuff we no longer need
+    delete $conf->{tests};
+    delete $conf->{priority};
+    delete $conf->{test_types};
+  }
+}
+
+sub trace_meta_dependencies {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+  $conf->{meta_dependencies} = { };
+
+  foreach my $name (keys %{$conf->{tests}}) {
+    next unless ($conf->{test_types}->{$name}
+                    == $Mail::SpamAssassin::Conf::TYPE_META_TESTS);
+
+    my $deps = [ ];
+    my $alreadydone = { };
+    $self->_meta_deps_recurse($conf, $name, $name, $deps, $alreadydone);
+    $conf->{meta_dependencies}->{$name} = join (' ', @{$deps});
+  }
+}
+
+sub _meta_deps_recurse {
+  my ($self, $conf, $toprule, $name, $deps, $alreadydone) = @_;
+
+  # Only do each rule once per top-level meta; avoid infinite recursion
+  return if $alreadydone->{$name};
+  $alreadydone->{$name} = 1;
+
+  # Obviously, don't trace empty or nonexistent rules
+  my $rule = $conf->{tests}->{$name};
+  return unless $rule;
+
+  # Lex the rule into tokens using a rather simple RE method ...
+  my $lexer = ARITH_EXPRESSION_LEXER;
+  my @tokens = ($rule =~ m/$lexer/g);
+
+  # Go through each token in the meta rule
+  foreach my $token (@tokens) {
+    # has to be an alpha+numeric token
+    next if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/);
+    # and has to be a rule name
+    next unless exists $conf->{tests}->{$token};
+
+    # add and recurse
+    push(@{$deps}, untaint_var($token));
+    $self->_meta_deps_recurse($conf, $toprule, $token, $deps, $alreadydone);
+  }
+}
+
+sub fix_priorities {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+
+  die unless $conf->{meta_dependencies};    # order requirement
+  my $pri = $conf->{priority};
+
+  # sort into priority order, lowest first -- this way we ensure that if we
+  # rearrange the pri of a rule early on, we cannot accidentally increase its
+  # priority later.
+  foreach my $rule (sort {
+            $pri->{$a} <=> $pri->{$b}
+          } keys %{$pri})
+  {
+    # we only need to worry about meta rules -- they are the
+    # only type of rules which depend on other rules
+    my $deps = $conf->{meta_dependencies}->{$rule};
+    next unless (defined $deps);
+
+    my $basepri = $pri->{$rule};
+    foreach my $dep (split ' ', $deps) {
+      my $deppri = $pri->{$dep};
+      if ($deppri > $basepri) {
+        dbg("rules: $rule (pri $basepri) requires $dep (pri $deppri): fixed");
+        $pri->{$dep} = $basepri;
+      }
+    }
+  }
+}
+
+sub find_dup_rules {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+
+  my %names_for_text;
+  my %dups;
+  while (my ($name, $text) = each %{$conf->{tests}}) {
+    my $type = $conf->{test_types}->{$name};
+
+    # skip eval and empty tests
+    next if ($type & 1) ||
+      ($type eq $Mail::SpamAssassin::Conf::TYPE_EMPTY_TESTS);
+
+    my $tf = ($conf->{tflags}->{$name}||''); $tf =~ s/\s+/ /gs;
+    # ensure similar, but differently-typed, rules are not marked as dups;
+    # take tflags into account too due to "tflags multiple"
+    $text = "$type\t$text\t$tf";
+
+    if (defined $names_for_text{$text}) {
+      $names_for_text{$text} .= " ".$name;
+      $dups{$text} = undef;     # found (at least) one
+    } else {
+      $names_for_text{$text} = $name;
+    }
+  }
+
+  foreach my $text (keys %dups) {
+    my $first;
+    my $first_pri;
+    my @names = sort {$a cmp $b} split(' ', $names_for_text{$text});
+    foreach my $name (@names) {
+      my $priority = $conf->{priority}->{$name} || 0;
+
+      if (!defined $first || $priority < $first_pri) {
+        $first_pri = $priority;
+        $first = $name;
+      }
+    }
+    # $first is now the earliest-occurring rule. mark others as dups
+
+    my @dups;
+    foreach my $name (@names) {
+      next if $name eq $first;
+      push @dups, $name;
+      delete $conf->{tests}->{$name};
+    }
+
+    dbg("rules: $first merged duplicates: ".join(' ', @dups));
+    $conf->{duplicate_rules}->{$first} = \@dups;
+  }
+}
+
+sub pack_eval_method {
+  my ($self, $function, $args, $name, $text) = @_;
+
+  my @args;
+  if (defined $args) {
+    # bug 4419: Parse quoted strings, unquoted alphanumerics/floats,
+    # unquoted IPv4 and IPv6 addresses, and unquoted common domain names.
+    # s// is used so that we can determine whether or not we successfully
+    # parsed ALL arguments.
+    local($1,$2,$3);
+    while ($args =~ s/^\s* (?: (['"]) (.*?) \1 | ( [\d\.:A-Za-z-]+? ) )
+                       \s* (?: , \s* | $ )//x) {
+      if (defined $2) {
+        push @args, $2;
+      }
+      else {
+        push @args, $3;
+      }
+    }
+  }
+
+  if ($args ne '') {
+    $self->lint_warn("syntax error (unparsable argument: $args) for eval function: $name: $text", $name);
+    return;
+  }
+
+  my $argstr = $function;
+  $argstr =~ s/\s+//gs;
+
+  if (@args > 0) {
+    $argstr .= ',' . join(', ',
+              map { my $s = $_; $s =~ s/\#/[HASH]/gs; 'q#' . $s . '#' } @args);
+  }
+  return ($argstr, \@args);
 }
 
 ###########################################################################
@@ -768,38 +1041,48 @@ sub lint_trusted_networks {
   my ($self) = @_;
   my $conf = $self->{conf};
 
-  my $nt = $conf->{trusted_networks};
-  my $ni = $conf->{internal_networks};
-
   # validate trusted_networks and internal_networks, bug 4760.
   # check that all internal_networks are listed in trusted_networks
-  # too.
+  # too.  do the same for msa_networks, but check msa_networks against
+  # internal_networks if trusted_networks aren't defined
+
+  my ($nt, $matching_against);
+  if ($conf->{trusted_networks_configured}) {
+    $nt = $conf->{trusted_networks};
+    $matching_against = 'trusted_networks';
+  } elsif ($conf->{internal_networks_configured}) {
+    $nt = $conf->{internal_networks};
+    $matching_against = 'internal_networks';
+  } else {
+    return;
+  }
+
+  foreach my $net_type ('internal_networks', 'msa_networks') {
+    next unless $conf->{"${net_type}_configured"};
+    next if $net_type eq $matching_against;
 
-  if ($ni->get_num_nets() > 0 && $nt->get_num_nets() > 0) {
     my $replace_nets;
-    my @valid_ni = ();
+    my @valid_net_list;
+    my $net_list = $conf->{$net_type};
 
-    foreach my $net (@{$ni->{nets}}) {
+    foreach my $net (@{$net_list->{nets}}) {
       # don't check to see if an excluded network is included - that's senseless
       if (!$net->{exclude} && !$nt->contains_net($net)) {
-        my $msg = "trusted_networks doesn't contain internal_networks entry '".
-                ($net->{as_string})."'\n";
+        my $msg = "$matching_against doesn't contain $net_type entry '".
+                  ($net->{as_string})."'";
 
-        if ($conf->{lint_rules}) {
-          warn $msg;
-          $conf->{errors}++;
-        }
+        $self->lint_warn($msg, undef);      # complain
         $replace_nets = 1;  # and omit it from the new internal set
       }
       else {
-        push @valid_ni, $net;
+        push @valid_net_list, $net;
       }
     }
 
     if ($replace_nets) {
       # something was invalid. replace the old nets list with a fixed version
       # (which may be empty)
-      $ni->{nets} = \@valid_ni;
+      $net_list->{nets} = \@valid_net_list;
     }
   }
 }
@@ -812,9 +1095,8 @@ sub add_test {
 
   # Don't allow invalid names ...
   if ($name !~ /^\D\w*$/) {
-    warn "config: error: rule '$name' has invalid characters ".
-	   "(not Alphanumeric + Underscore + starting with a non-digit)\n";
-    $conf->{errors}++;
+    $self->lint_warn("config: error: rule '$name' has invalid characters ".
+	   "(not Alphanumeric + Underscore + starting with a non-digit)\n", $name);
     return;
   }
 
@@ -822,18 +1104,16 @@ sub add_test {
   # characters throw warnings).  Check this separately from the above
   # pattern to avoid vague error messages.
   if (length $name > 200) {
-    warn "config: error: rule '$name' is way too long ".
-	   "(recommended maximum length is 22 characters)\n";
-    $conf->{errors}++;
+    $self->lint_warn("config: error: rule '$name' is way too long ".
+	   "(recommended maximum length is 22 characters)\n", $name);
     return;
   }
 
   # Warn about, but use, long rule names during --lint
   if ($conf->{lint_rules}) {
     if (length($name) > 50 && $name !~ /^__/ && $name !~ /^T_/) {
-      warn "config: warning: rule name '$name' is over 50 chars ".
-	     "(recommended maximum length is 22 characters)\n";
-      $conf->{errors}++;
+      $self->lint_warn("config: warning: rule name '$name' is over 50 chars ".
+	     "(recommended maximum length is 22 characters)\n", $name);
     }
   }
 
@@ -847,9 +1127,13 @@ sub add_test {
   }
   if ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_TESTS)
   {
-    my ($pat) = ($text =~ /^\s*\S+\s*(?:\=|\!)\~\s*(\S.*?\S)\s*$/);
-    $pat =~ s/\s+\[if-unset:\s+(.+)\]\s*$//;
-    return unless $self->is_delimited_regexp_valid($name, $pat);
+    if ($text =~ /^!?defined\([A-Za-z][A-Za-z0-9-]*\)$/) {
+      # fine, implements 'exists:'
+    } else {
+      my ($pat) = ($text =~ /^\s*\S+\s*(?:\=|\!)\~\s*(\S.*?\S)\s*$/);
+      if ($pat) { $pat =~ s/\s+\[if-unset:\s+(.+)\]\s*$//; }
+      return unless $self->is_delimited_regexp_valid($name, $pat);
+    }
   }
   elsif ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS)
   {
@@ -858,12 +1142,38 @@ sub add_test {
 
   $conf->{tests}->{$name} = $text;
   $conf->{test_types}->{$name} = $type;
-  $conf->{tflags}->{$name} ||= '';
+  if ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS) {
+    $conf->{priority}->{$name} ||= 500;
+  }
+  else {
+    $conf->{priority}->{$name} ||= 0;
+  }
   $conf->{priority}->{$name} ||= 0;
   $conf->{source_file}->{$name} = $self->{currentfile};
 
+  if ($conf->{main}->{keep_config_parsing_metadata}) {
+    $conf->{if_stack}->{$name} = $self->get_if_stack_as_string();
+
+    if ($self->{file_scoped_attrs}->{testrules}) {
+      $conf->{testrules}->{$name} = 1;   # used in build/mkupdates/listpromotable
+    }
+  }
+
+  # if we found this rule in a user_prefs file, it's a user rule -- note that
+  # we may need to recompile the rule code for this type (if they've already
+  # been compiled, e.g. in spamd).
+  #
+  # Note: the want_rebuild_for_type 'flag' is actually a counter; it is decremented
+  # after each scan.  This ensures that we always recompile at least once more;
+  # once to *define* the rule, and once afterwards to *undefine* the rule in the
+  # compiled ruleset again.
+  #
+  # If two consecutive scans use user rules, that's ok -- the second one will
+  # reset the counter, and we'll still recompile just once afterwards to undefine
+  # the rule again.
+  #
   if ($self->{scoresonly}) {
-    $conf->{user_rules_to_compile}->{$type} = 1;
+    $conf->{want_rebuild_for_type}->{$type} = 2;
     $conf->{user_defined_rules}->{$name} = 1;
   }
 }
@@ -885,12 +1195,15 @@ sub is_meta_valid {
   my ($self, $name, $rule) = @_;
 
   my $meta = '';
+  $rule = untaint_var($rule);  # must be careful below
 
   # Lex the rule into tokens using a rather simple RE method ...
   my $lexer = ARITH_EXPRESSION_LEXER;
   my @tokens = ($rule =~ m/$lexer/g);
   if (length($name) == 1) {
-    print "$name $_\n " for @tokens;
+    for (@tokens) {
+      print "$name $_\n "  or die "Error writing token: $!";
+    }
   }
   # Go through each token in the meta rule
   foreach my $token (@tokens) {
@@ -908,22 +1221,19 @@ sub is_meta_valid {
   if (eval $evalstr) {
     return 1;
   }
-  if ($@) {
-    my $err = $@;
-    $err =~ s/\s+(?:at|near)\b.*//s;
-    $err =~ s/Illegal division by zero/division by zero possible/i;
-    warn "config: invalid expression for rule $name: \"$rule\": $err\n";
-    $self->{conf}->{errors}++;
-    return 0;
-  }
+  my $err = $@ ne '' ? $@ : "errno=$!";  chomp $err;
+  $err =~ s/\s+(?:at|near)\b.*//s;
+  $err =~ s/Illegal division by zero/division by zero possible/i;
+  $self->lint_warn("config: invalid expression for rule $name: \"$rule\": $err\n", $name);
+  return 0;
 }
 
 sub is_delimited_regexp_valid {
   my ($self, $name, $re) = @_;
 
-  unless ($re =~ /^\s*m?(\W).*(?:\1|>|}|\)|\])[a-z]*\s*$/) {
-    warn "config: invalid regexp for rule $name: $re: missing or invalid delimiters\n";
-    $self->{conf}->{errors}++;
+  if (!$re || $re !~ /^\s*m?(\W).*(?:\1|>|}|\)|\])[a-z]*\s*$/) {
+    $re ||= '';
+    $self->lint_warn("config: invalid regexp for rule $name: $re: missing or invalid delimiters\n", $name);
     return 0;
   }
   return $self->is_regexp_valid($name, $re);
@@ -938,6 +1248,7 @@ sub is_regexp_valid {
   my $origre = $re;
   my $safere = $re;
   my $mods = '';
+  local ($1,$2);
   if ($re =~ s/^m{//) {
     $re =~ s/}([a-z]*)$//; $mods = $1;
   }
@@ -957,6 +1268,21 @@ sub is_regexp_valid {
     $safere = "m#".$re."#";
   }
 
+  if ($self->{conf}->{lint_rules} ||
+      $self->{conf}->{ignore_always_matching_regexps})
+  {
+    my $msg = $self->is_always_matching_regexp($name, $re);
+
+    if (defined $msg) {
+      if ($self->{conf}->{lint_rules}) {
+        $self->lint_warn($msg, $name);
+      } else {
+        warn $msg;
+        return 0;
+      }
+    }
+  }
+
   # now prepend the modifiers, in order to check if they're valid
   if ($mods) {
     $re = "(?".$mods.")".$re;
@@ -969,14 +1295,30 @@ sub is_regexp_valid {
   if (eval { ("" =~ m#${re}#); 1; }) {
     return 1;
   }
-
-  my $err = $@;
+  my $err = $@ ne '' ? $@ : "errno=$!";  chomp $err;
   $err =~ s/ at .*? line \d.*$//;
-  warn "config: invalid regexp for rule $name: $origre: $err\n";
-  $self->{conf}->{errors}++;
+  $self->lint_warn("config: invalid regexp for rule $name: $origre: $err\n", $name);
   return 0;
 }
 
+# check the pattern for some basic errors, and warn if found
+sub is_always_matching_regexp {
+  my ($self, $name, $re) = @_;
+
+  if ($re =~ /(?<!\\)\|\|/) {
+    return "config: regexp for rule $name always matches due to '||'";
+  }
+  elsif ($re =~ /^\|/) {
+    return "config: regexp for rule $name always matches due to " .
+      "pattern starting with '|'";
+  }
+  elsif ($re =~ /\|(?<!\\\|)$/) {
+    return "config: regexp for rule $name always matches due to " .
+      "pattern ending with '|'";
+  }
+  return;
+}
+
 ###########################################################################
 
 sub add_to_addrlist {
@@ -1036,14 +1378,52 @@ sub remove_from_addrlist_rcvd {
 sub fix_path_relative_to_current_file {
   my ($self, $path) = @_;
 
+  # the path may be specified as "~/foo", so deal with that
+  $path = $self->{conf}->{main}->sed_path($path);
+
   if (!File::Spec->file_name_is_absolute ($path)) {
     my ($vol, $dirs, $file) = File::Spec->splitpath ($self->{currentfile});
     $path = File::Spec->catpath ($vol, $dirs, $path);
-    dbg("plugin: fixed relative path: $path");
+    dbg("config: fixed relative path: $path");
   }
   return $path;
 }
 
 ###########################################################################
 
+sub lint_warn {
+  my ($self, $msg, $rule, $iserror) = @_;
+
+  if (!defined $iserror) { $iserror = 1; }
+
+  if ($self->{conf}->{main}->{lint_callback}) {
+    $self->{conf}->{main}->{lint_callback}->(
+          msg => $msg,
+          rule => $rule,
+          iserror => $iserror
+        );
+  }
+  elsif ($self->{conf}->{lint_rules}) {
+    warn $msg."\n";
+  }
+  else {
+    info($msg);
+  }
+
+  if ($iserror) {
+    $self->{conf}->{errors}++;
+  }
+}
+
+###########################################################################
+
+sub get_if_stack_as_string {
+  my ($self) = @_;
+  return join ' ', map {
+    $_->{conditional}
+  } @{$self->{if_stack}};
+}
+
+###########################################################################
+
 1;
