@@ -22,6 +22,7 @@ package Mail::SpamAssassin::SubProcBackChannel;
 use strict;
 use warnings;
 use bytes;
+use re 'taint';
 
 use IO::Socket;
 use Mail::SpamAssassin::Util;
@@ -86,11 +87,12 @@ sub setup_backchannel_parent_post_fork {
 
   my $fh = $self->{latest_kid_fh};
 
-  close $self->{parent};    # because it's us!
+  close $self->{parent}    # because it's us!
+    or die "backchannel: error closing parent side of the pipe: $!";
 
   # disable caching for parent<->child relations
   my ($old) = select($fh);
-  $|++;
+  $| = 1;   # turn off buffering
   select($old);
 
   $self->{kids}->{$pid} = $fh;
@@ -99,11 +101,24 @@ sub setup_backchannel_parent_post_fork {
 
 sub add_to_selector {
   my ($self, $fh) = @_;
+  if (!defined $fh) {
+    warn "undef fh in add_to_selector"; return;
+  }
   my $fno = fileno($fh);
   $self->{fileno_to_fh}->{$fno} = $fh;
   vec (${$self->{selector}}, $fno, 1) = 1;
 }
 
+sub remove_from_selector {
+  my ($self, $fh) = @_;
+  if (!defined $fh) {
+    warn "undef fh in remove_from_selector"; return;
+  }
+  my $fno = fileno($fh);
+  delete $self->{fileno_to_fh}->{$fno};
+  vec (${$self->{selector}}, $fno, 1) = 0;
+}
+
 sub select_vec_to_fh_list {
   my ($self, $vec) = @_;
   my $i = -1;
@@ -137,7 +152,8 @@ sub delete_socket_for_child {
 sub setup_backchannel_child_post_fork {
   my ($self) = @_;
 
-  close $self->{latest_kid_fh}; # because it's us!
+  close $self->{latest_kid_fh}  # because it's us!
+    or die "backchannel: error closing child side of the pipe: $!";
 
   my $old = select($self->{parent});
   $| = 1;   # print to parent by default, turn off buffering
